// src/entry.rs

/// This is the first function called when the system boots. In this function,
/// we set up important control status registers (CSR) such as `mstatus` and
/// `satp`. The global, stack, and thread pointers are also configured here.
#[no_mangle]
#[link_section = ".text.init"]
unsafe extern "C" fn entry() {
    use core::arch::asm;
    use riscv::register::{mhartid, mstatus, satp, sstatus};

    // Global pointer `gp`
    //
    // Push current option stack to temporarily disable relaxation, load
    // _global_pointer symbol (provided by linker), then pop option stack.
    //
    // Important to keep relaxation off for this instruction, so that
    // the instruction is emitted as:
    //     1:
    //         auipc gp, %pcrel_hi(_global_pointer)
    //         addi gp, gp, %pcrel_lo(1b)"
    // instead of:
    //         mv gp, gp
    asm!(
        ".option push",
        ".option norelax",
        "la gp, _global_pointer",
        ".option pop",
    );

    // Stack pointer `sp`
    //
    // Set up the stack pointer for each hart, since each hart will have its own stack.
    asm!("la sp, _stack_end");

    // Thread pointer `tp`
    //
    // Set up the thread pointer, which will hold the hart id for each hart.
    write_tp(&mhartid::read());

    satp::write(0);

    mstatus::set_mpp(mstatus::MPP::Machine);
    mstatus::set_fs(sstatus::FS::Initial);

    asm!(
        "csrw mie, x0",

        "la t1, {kinit}",
        "csrw mepc, t1",

        "la ra, 2f",
        "mret",
    "2:",
        kinit = sym kinit,
    );
}

// TODO ...then kinit

#[no_mangle]
extern "C" fn kinit() {
    use crate::uart;
    use core::arch::asm;
    uart::Device::new(0x1000_0000);

    if read_tp() == 0 {
        // Clear bss and set stack pointer
        // Only hart 0 does this
        unsafe {
            asm!(
                "la t0, _bss_start",
                "la t1, _bss_end",
                "bgeu t0, t1, 2f",
            "1:",
                "sw x0, (t0)", // Store zero as word at t0
                "addi t0, t0, 4",
                "bne t0, t1, 1b",
            "2:",
            );
        }
    }
}

#[no_mangle]
extern "C" fn kinit_hart() {}

#[inline]
fn read_tp() -> usize {
    use core::arch::asm;
    let id: usize;
    unsafe {
        asm!(
            "mv {id}, tp",
            id = out(reg) id,
        );
    }
    id
}

#[inline]
fn write_tp(id: &usize) {
    use core::arch::asm;
    unsafe {
        asm!(
            "csrw mhartid, {id}",
            id = in(reg) id,
        );
    }
}
